const int ROOTNODE=0;					// for more descriptive parameters - send to head node in cluster 
typedef double * PDOUBLE;

//-- message codes for socket transfers
const int SOCKMSG_EXIT		= 2048;
const int SOCKMSG_COMPLETE	= 2049;
const int SOCKMSG_SIMULATE	= 2050;
const int SOCKMSG_SIMRESULTS= 2051;

#ifdef MPIPROG
#define WSAGetLastError() ((int)0)
#endif

//------------------------------------------------------------------- 
// Sim parameters message 
class CSimParameters { 
	public : 
		int m_iSegments; 
		int m_iPoints; 
		int m_iSeed; 
		int m_iChannels;	
		int m_iSampleRate; 
		int m_iStates; 
		double * m_adAmp; 
		double * m_adSD;
		double * m_adP0; 
		double ** m_aQ;

		CSimParameters() { 
			m_adAmp=NULL;
			m_adSD=NULL;
			m_adP0=NULL;
			m_aQ=NULL;
			}
			
		~CSimParameters() { 
			if( m_adAmp!=NULL )
				delete [] m_adAmp;
			if( m_adSD!=NULL )
				delete [] m_adSD;			
			if( m_adP0!=NULL )
				delete [] m_adP0;
			if( m_aQ!=NULL ) { 
				for( int i=0; i<m_iStates; ++i ) 
					if( m_aQ[i]!=NULL ) 
						delete [] m_aQ[i];
				delete [] m_aQ;
				}
			}

		int BufferSize( int iStates ) { 
			return 24+24*iStates+8*iStates*iStates;
			}
			
		bool FromBuffer( int iBytes, BYTE * pBuf ) { 
			m_iSegments = *(int *)(pBuf+0);
			m_iPoints = *(int *)(pBuf+4);
			m_iSeed = *(int *)(pBuf+8);
			m_iChannels = *(int *)(pBuf+12);
			m_iSampleRate = *(int *)(pBuf+16);
			m_iStates = *(int *)(pBuf+20);
			if( BufferSize(m_iStates) != iBytes )
				return false;
			if( (m_adAmp=new double[m_iStates])==NULL
				|| (m_adSD=new double[m_iStates])==NULL
				|| (m_adP0=new double[m_iStates])==NULL
				|| (m_aQ=new PDOUBLE[m_iStates])==NULL )  
				return false;
			for( int iState=0; iState<m_iStates; ++iState )
				if( (m_aQ[iState]=new double[m_iStates]) == NULL ) 
					return false; 
			double * pData=(double *)(pBuf+24);
			for( int iState=0; iState<m_iStates; ++iState ) {
				m_adAmp[iState] = *(pData);
				m_adSD[iState] = *(pData+1);
				m_adP0[iState] = *(pData+2);
				pData+=3;
				}
			for( int iState=0; iState<m_iStates; ++iState )
				for( int iCol=0; iCol<m_iStates; ++iCol ) {
					m_aQ[iState][iCol]=*(double *)pData;
					pData++;
					}
			return true;
			}
			
		bool ToBuffer( BYTE * pBuf ) { 
			*(int *)(pBuf+ 0)=m_iSegments;
			*(int *)(pBuf+ 4)=m_iPoints;
			*(int *)(pBuf+ 8)=m_iSeed;
			*(int *)(pBuf+12)=m_iChannels;
			*(int *)(pBuf+16)=m_iSampleRate;
			*(int *)(pBuf+20)=m_iStates;
			double * pData=(double *)(pBuf+24);
			for( int iState=0; iState<m_iStates; ++iState ) {
				*(pData)=m_adAmp[iState];
				*(pData+1)=m_adSD[iState];
				*(pData+2)=m_adP0[iState];
				pData+=3;
				}
			for( int iState=0; iState<m_iStates; ++iState )
				for( int iCol=0; iCol<m_iStates; ++iCol ) {
					*pData=m_aQ[iState][iCol];
					pData++;
					}
			return true;
			}
	
		#ifdef MPIPROG
		bool Send( int iNode ) { 
			int iSize = BufferSize(m_iStates);
			BYTE * pBuf = new BYTE[iSize];
			if( ! ToBuffer( pBuf ) ) 
				return false;
			MPI::COMM_WORLD.Send( pBuf, iSize, MPI::BYTE, iNode, MPIMSG_SIMPARAMS );
			#ifdef _DEBUG
				printf("Parameters to node=%d Size=%d\n", iNode, iSize );
			#endif
			delete [] pBuf;
			return true;
			}
		#endif
			
		#ifdef MPIPROG
		bool Recv( int iBytes ) { 
			if( iBytes < 24+24*2+8*2*2 )		// at least 2 states 
				return false;
			BYTE * pBuf = new BYTE[iBytes];
			if( pBuf==NULL ) 
				return false;
			MPI::COMM_WORLD.Recv( pBuf, iBytes, MPI::BYTE, ROOTNODE, MPIMSG_SIMPARAMS );
			#ifdef _DEBUG
				printf("Parameters from root Size=%d\n", iBytes );
			#endif
			if( ! FromBuffer( iBytes, pBuf ) ) {
				delete [] pBuf;
				return false;
				}
			delete [] pBuf;
			return true;
			}
		#endif

		bool Allocate( int iStates, double * adAmp, double * adSD, double * adP0, double ** aQ ) { 
			m_iStates=iStates;
			if( (m_adAmp = new double[iStates]) == NULL 
				|| (m_adSD = new double[iStates]) == NULL 
				|| (m_adP0 = new double[iStates]) == NULL 
				|| (m_aQ = new PDOUBLE[iStates]) == NULL )
					return false;
			for( int iState=0; iState<iStates; ++iState ) 
				if( (m_aQ[iState]=new double[iStates])==NULL ) 
					return false;
			for( int i=0; i<iStates; ++i ) {
				if( adAmp != NULL ) 
					m_adAmp[i]=adAmp[i];
				if( adSD != NULL ) 
					m_adSD[i]=adSD[i];
				if( adP0 != NULL ) 
					m_adP0[i]=adP0[i];
				if( aQ != NULL ) 
					for( int j=0; j<iStates; ++j )
						m_aQ[i][j]=aQ[i][j];
				}
			return true;
			}	
							
		};

//------------------------------------------------------------------- 
// Sim results message 
class CSimResults { 
	public : 
		int m_iSegment;
		int m_iDwells;
		int * m_aiPoints;
		int * m_aiStates;
		int m_iPoints;
		double * m_adData;

		int m_iDwellsAllocated;

		CSimResults() { 
			m_aiPoints=NULL;
			m_aiStates=NULL;
			m_adData=NULL;
			m_iDwellsAllocated=0;
			}
			
		~CSimResults() { 
			if( m_aiPoints!=NULL )
				delete [] m_aiPoints;
			if( m_aiStates!=NULL )
				delete [] m_aiStates;
			if( m_adData!=NULL )
				delete [] m_adData;
			}

		bool AllocateData( int iPoints ) { 
			m_iDwells=0;
			return (m_adData = new double[m_iPoints=iPoints]) != NULL ;
			}
					
		int BufferSize( int iDwells, int iPoints ) { 
			return 12+8*iDwells+8*iPoints;
			}
			
		bool FromBuffer( int iBytes, BYTE * pBuf ) { 
			m_iSegment = *(int *)(pBuf+0);
			m_iDwells = *(int *)(pBuf+4);
			m_iPoints = *(int *)(pBuf+8);

			if( BufferSize(m_iDwells,m_iPoints) != iBytes )
				return false;
			if( (m_aiPoints=new int[m_iDwells])==NULL
				|| (m_aiStates=new int[m_iDwells])==NULL
				|| (m_adData=new double[m_iPoints])==NULL ) 
				return false;

			int * piData=(int *)(pBuf+12);
			for( int iDwell=0; iDwell<m_iDwells; iDwell++ ) { 
				m_aiPoints[iDwell]=*(piData+0);
				m_aiStates[iDwell]=*(piData+1);
				piData+=2;
				}
			double * pdData=(double *)piData;
			for( int iPoint=0; iPoint<m_iPoints; ++iPoint )
				m_adData[iPoint]=*(pdData++);
			return true;
			}
			
		bool ToBuffer( BYTE * pBuf ) { 
			*(int *)(pBuf+0)=m_iSegment;
			*(int *)(pBuf+4)=m_iDwells;
			*(int *)(pBuf+8)=m_iPoints;
			int * piData=(int *)(pBuf+12);
			for( int iDwell=0; iDwell<m_iDwells; iDwell++ ) { 
				*(piData+0)=m_aiPoints[iDwell];
				*(piData+1)=m_aiStates[iDwell];
				piData+=2;
				}
			double * pdData=(double *)piData;
			for( int iPoint=0; iPoint<m_iPoints; ++iPoint )
				*(pdData++)=m_adData[iPoint];
			return true;
			}
	
		#ifdef MPIPROG
		bool Send( ) { 
			int iSize = BufferSize( m_iDwells, m_iPoints);
			BYTE * pBuf = new BYTE[iSize];
			if( ! ToBuffer( pBuf ) ) 
				return false;
			MPI::COMM_WORLD.Send( pBuf, iSize, MPI::BYTE, ROOTNODE, MPIMSG_SIMRESULTS );
			#ifdef _DEBUG
				printf("Results to root Size=%d\n", iSize );
			#endif
			delete [] pBuf;
			return true;
			}
		#endif
			
		#ifdef MPIPROG
		bool Recv( int iNode, int iBytes, BYTE * pBuf ) { 
			MPI::COMM_WORLD.Recv( pBuf, iBytes, MPI::BYTE, iNode, MPIMSG_SIMRESULTS );
			#ifdef _DEBUG
				printf("Results from node=%d Size=%d ", iNode, iBytes );
			#endif
			return true;
			}
		#endif

		#ifdef MPIPROG
		bool Recv( int iNode, int iBytes ) { 
			if( iBytes < 12+8+8 )		// at least 1 dwell 
				return false;
			BYTE * pBuf = new BYTE[iBytes];
			if( pBuf==NULL ) 
				return false;
			MPI::COMM_WORLD.Recv( pBuf, iBytes, MPI::BYTE, iNode, MPIMSG_SIMRESULTS );
			#ifdef _DEBUG
				printf("Results to root from node=%d Size=%d\n", iNode, iBytes );
			#endif
			if( ! FromBuffer( iBytes, pBuf ) ) {
				delete [] pBuf;
				return false;
				}
			delete [] pBuf;
			return true;
			}
		#endif

		bool AddDwell( int iPoints, int iState ) { 
			if( m_iDwells==0 ) { 
				m_iDwellsAllocated=65000;		// ~ 1mb
				m_aiPoints = new int[m_iDwellsAllocated];
				m_aiStates = new int[m_iDwellsAllocated];
				if( m_aiPoints==NULL || m_aiStates==NULL ) { 
					delete [] m_aiPoints;
					m_iDwellsAllocated=0;
					return false;
					}
				}
			else if( m_iDwells>=m_iDwellsAllocated ) {
				//-- allocate new 
				int * aiPoints=NULL, * aiStates;
				if( ( aiPoints=new int[2*m_iDwellsAllocated] ) == NULL 
					|| (aiStates = new int[2*m_iDwellsAllocated]) == NULL ) { 
					delete [] aiPoints;
					return false;
					}
				//-- copy old to new
				for( int iDwell=0; iDwell<m_iDwells; ++iDwell ) { 
					aiPoints[iDwell]=m_aiPoints[iDwell];
					aiStates[iDwell]=m_aiStates[iDwell];
					}
				//-- delete old
				delete [] m_aiPoints;
				m_aiPoints=aiPoints;
				delete [] m_aiStates;
				m_aiStates=aiStates;
				}
			
			m_aiStates[m_iDwells]=iState;
			m_aiPoints[m_iDwells]=iPoints;
			++m_iDwells;
			}

		void SetPoint( bool bIncrement, int iPoint, double dValue ) {
			if( bIncrement ) 
				m_adData[iPoint] += dValue; 
			else 
				m_adData[iPoint] = dValue; 
			}
				
		};

class CSocket { 
	public : 
		SOCKET m_sockListen;
		sockaddr_in m_addrListen;
		SOCKET m_sockClient;
			
		//-- prepare socket for connections (listen)	
		bool Setup() { 
			//-- create the socket 
			m_sockListen = socket( AF_INET, SOCK_STREAM, 0 );
			if( m_sockListen == INVALID_SOCKET )
				return false;

			//-- build address structure	
			memset( &m_addrListen, 0, sizeof( m_addrListen ));
			m_addrListen.sin_family = AF_INET;
			m_addrListen.sin_port = htons(64237);
			m_addrListen.sin_addr.s_addr = INADDR_ANY;

			//-- bind listening port / address
			if( bind( m_sockListen, (struct sockaddr*) &m_addrListen, sizeof( m_addrListen ) ) != 0 )
				return false;
		
			//-- Set socket to listening mode 
			if( listen( m_sockListen, SOMAXCONN ) != 0 ) 
				return false;
			return true;
			}
	
		//-- check if there is an incoming request
		bool Poll( int iSec ) { 
			// construct 0 sec time limit 
			timeval tv;
			tv.tv_sec=iSec;
			tv.tv_usec=0;
			
			// construct socket read file descriptor set 			
			fd_set fdsock;
			FD_ZERO( &fdsock );
			FD_SET( m_sockListen, &fdsock);
			
			int iSelect = select(FD_SETSIZE, &fdsock, NULL, NULL, &tv);
			
			#ifdef _DEBUG
			printf("Socket Select completed %d\n", iSelect );
			#endif 
			
			return iSelect!=0;
			}
		
		//-- get a buffer of data of known size from the socket.
		bool GetBuffer( int iLen, BYTE * pData ) { 
			int iRecvLen, iRecv = 0; 
			while( iRecv<iLen ) {
				iRecvLen = recv( m_sockClient, (char *)(pData+iRecv), iLen-iRecv, 0 );
				if( iRecvLen==SOCKET_ERROR || iRecvLen==0 ) {
					#ifdef _DEBUG
						printf("Socket error on receive len=%d-%d err=%d\n", iLen, iRecv, WSAGetLastError() );
					#endif
					closesocket( m_sockClient );
					return false;
					}
				#ifdef _DEBUG
				else if( iRecv+iRecvLen==iLen )
					printf( "Socket GetRequest len=%d completed.\n", iRecv );
				else 
					printf( "Socket GetRequest len=%d got=%d had=%d\n", iLen, iRecvLen, iRecv );
				#endif
				iRecv += iRecvLen;
				}
			return true;
			}
				
		//-- Get the code and length for next message 
		bool GetCodeAndLength( int * piCode, int * piBytes ) { 
			//-- get code and length
			BYTE buf[8]; 
			if( ! GetBuffer( 8, buf ) ) 
				return false;
			*piCode=*(int *)buf;
			*piBytes=*(int *)(buf+4);
			#ifdef _DEBUG
			printf( "Socket Get Code(%d) and Length(%d)\n", *piCode, *piBytes );
			#endif
			return true;
			}

		//-- accept the request and open the client socket
		// For this app, get 8 bytes of data from the socket : 4=code, 4=size 
		// Then process the remaining appropriately.  
		bool GetRequest( int * piCode, int * piBytes, BYTE ** pData ) { 
			sockaddr_in addrClient;				// socket to contain client address of connect request
			socklen_t iAddrLen = sizeof(sockaddr_in);	
			//-- accept a connection 
			m_sockClient = accept( m_sockListen, (struct sockaddr*)&addrClient, &iAddrLen);
			if( m_sockClient == INVALID_SOCKET ) 
				return false;
				
			//-- get code and length 
			if( ! GetCodeAndLength( piCode, piBytes ) ) 
				return false;
				
			//-- Allocate buffer and fill it 
			*pData = new BYTE[*piBytes];
			if( *pData==NULL ) {
				#ifdef _DEBUG
					printf("Socket error allocating buffer size=%d\n", *piBytes );
				#endif
				return false;
				}
				
			return GetBuffer( *piBytes, *pData );
			}
		
		//-- send a buffer to the socket 
		bool SendBuffer( int iLen, BYTE * pData ) { 
			int iSentLen=send( m_sockClient, (char *)pData, iLen, 0 );
			if( iLen!=iSentLen ) {
				#ifdef _DEBUG
					printf("Socket SendBuffer error len=%d sent=%d error=%d\n", iLen, iSentLen, WSAGetLastError() );
				#endif
				return false;
				}
			return true;
			}

		//-- Send Code and length to socket 
		bool SendCodeAndLength( int iCode, int iLen ) { 
			char buffer[8];
			*(int *)buffer=iCode;
			*(int *)(buffer+4)=iLen;
			return SendBuffer( 8, (BYTE *)buffer );
			}
						
		//-- Response 
		// Send 8 bytes back to client ( code, length ) followed by response data.
		bool SendResponse( int iCode, int iSendLen, BYTE * pData ) { 
			return SendCodeAndLength( iCode, iSendLen ) && ( iSendLen==0 || SendBuffer( iSendLen, pData ) );
			}
			
		void CloseClient( ) { 
			//--Close data connection
			closesocket( m_sockClient );
			}

	};
